@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import argparse
 import ast
 import collections
@@ -14,11 +16,8 @@ from typing import Generator
 from typing import Iterable
 from typing import List
 from typing import NamedTuple
-from typing import Optional
 from typing import Sequence
-from typing import Set
 from typing import Tuple
-from typing import Union
 
 from aspy.refactor_imports.import_obj import AbstractImportObj
 from aspy.refactor_imports.import_obj import import_obj_from_str
@@ -43,9 +42,9 @@ Step = Callable[[Any], List[CodePartition]]
 
 class TopLevelImportVisitor(ast.NodeVisitor):
     def __init__(self) -> None:
-        self.top_level_import_line_numbers: List[int] = []
+        self.top_level_import_line_numbers: list[int] = []
 
-    def _visit_import(self, node: Union[ast.Import, ast.ImportFrom]) -> None:
+    def _visit_import(self, node: ast.Import | ast.ImportFrom) -> None:
         # If it's indented, we don't really care about the import.
         if node.col_offset == 0:
             self.top_level_import_line_numbers.append(node.lineno)
@@ -53,7 +52,7 @@ class TopLevelImportVisitor(ast.NodeVisitor):
     visit_Import = visit_ImportFrom = _visit_import
 
 
-def get_line_offsets_by_line_no(src: str) -> List[int]:
+def get_line_offsets_by_line_no(src: str) -> list[int]:
     # Padded so we can index with line number
     offsets = [0, 0]
     for line in src.splitlines(True):
@@ -65,7 +64,7 @@ def _partitions_to_src(partitions: Iterable[CodePartition]) -> str:
     return ''.join(part.src for part in partitions)
 
 
-def partition_source(src: str) -> List[CodePartition]:
+def partition_source(src: str) -> list[CodePartition]:
     """Partitions source into a list of `CodePartition`s for import
     refactoring.
     """
@@ -143,7 +142,7 @@ def partition_source(src: str) -> List[CodePartition]:
 
 def combine_trailing_code_chunks(
         partitions: Iterable[CodePartition],
-) -> List[CodePartition]:
+) -> list[CodePartition]:
     chunks = list(partitions)
 
     NON_COMBINABLE = (CodeType.IMPORT, CodeType.PRE_IMPORT_CODE)
@@ -158,7 +157,7 @@ def combine_trailing_code_chunks(
 
 def separate_comma_imports(
         partitions: Iterable[CodePartition],
-) -> List[CodePartition]:
+) -> list[CodePartition]:
     """Turns `import a, b` into `import a` and `import b`"""
     def _inner() -> Generator[CodePartition, None, None]:
         for partition in partitions:
@@ -179,8 +178,8 @@ def separate_comma_imports(
 
 def add_imports(
         partitions: Iterable[CodePartition],
-        to_add: Tuple[str, ...] = (),
-) -> List[CodePartition]:
+        to_add: tuple[str, ...] = (),
+) -> list[CodePartition]:
     partitions = list(partitions)
     if not _partitions_to_src(partitions).strip():
         return partitions
@@ -197,9 +196,9 @@ def add_imports(
 
 def remove_imports(
         partitions: Iterable[CodePartition],
-        to_remove: Tuple[str, ...] = (),
-) -> List[CodePartition]:
-    to_remove_imports: Set[AbstractImportObj] = set()
+        to_remove: tuple[str, ...] = (),
+) -> list[CodePartition]:
+    to_remove_imports: set[AbstractImportObj] = set()
     for s in to_remove:
         to_remove_imports.update(import_obj_from_str(s).split_imports())
 
@@ -214,14 +213,14 @@ def remove_imports(
     return list(_inner())
 
 
-def _mod_startswith(mod_parts: List[str], prefix_parts: List[str]) -> bool:
+def _mod_startswith(mod_parts: list[str], prefix_parts: list[str]) -> bool:
     return mod_parts[:len(prefix_parts)] == prefix_parts
 
 
 def replace_imports(
         partitions: Iterable[CodePartition],
         to_replace: Iterable[ImportToReplace] = (),
-) -> List[CodePartition]:
+) -> list[CodePartition]:
     def _inner() -> Generator[CodePartition, None, None]:
         for partition in partitions:
             if partition.code_type is CodeType.IMPORT:
@@ -275,9 +274,9 @@ def _module_to_base_modules(s: str) -> Generator[str, None, None]:
 
 def remove_duplicated_imports(
         partitions: Iterable[CodePartition],
-) -> List[CodePartition]:
-    seen: Set[AbstractImportObj] = set()
-    seen_module_names: Set[str] = set()
+) -> list[CodePartition]:
+    seen: set[AbstractImportObj] = set()
+    seen_module_names: set[str] = set()
     without_exact_duplicates = []
 
     for partition in partitions:
@@ -318,11 +317,11 @@ def apply_import_sorting(
         separate_relative: bool = False,
         separate_from_import: bool = False,
         **sort_kwargs: Any,
-) -> List[CodePartition]:
-    pre_import_code: List[CodePartition] = []
-    imports: List[CodePartition] = []
-    trash: List[CodePartition] = []
-    rest: List[CodePartition] = []
+) -> list[CodePartition]:
+    pre_import_code: list[CodePartition] = []
+    imports: list[CodePartition] = []
+    trash: list[CodePartition] = []
+    rest: list[CodePartition] = []
     for partition in partitions:
         {
             CodeType.PRE_IMPORT_CODE: pre_import_code,
@@ -348,7 +347,7 @@ def apply_import_sorting(
     relative_imports = []
 
     def _import_type_switches(
-            last_import_obj: Optional[AbstractImportObj],
+            last_import_obj: AbstractImportObj | None,
             import_obj: AbstractImportObj,
     ) -> bool:
         """Returns True if separate_from_import is True and  `import_obj` is
@@ -384,7 +383,11 @@ def apply_import_sorting(
         if last_import_obj is not None:
             new_imports.append(CodePartition(CodeType.NON_CODE, '\n'))
 
-    if relative_imports:
+    # There is another edge case if --separate-relative is passed while all the
+    # imports are relative. In that case we don't want an empty new line at the
+    # beginning of the block. We should insert the new line only if there are
+    # additional imports.  See #134
+    if relative_imports and len(relative_imports) != len(imports):
         relative_imports.insert(0, CodePartition(CodeType.NON_CODE, '\n'))
     # XXX: I want something like [x].join(...) (like str join) but for now
     # this works
@@ -405,8 +408,8 @@ def apply_import_sorting(
 
 
 def _get_steps(
-        imports_to_add: Tuple[str, ...],
-        imports_to_remove: Tuple[str, ...],
+        imports_to_add: tuple[str, ...],
+        imports_to_remove: tuple[str, ...],
         imports_to_replace: Iterable[ImportToReplace],
         **sort_kwargs: Any,
 ) -> Generator[Step, None, None]:
@@ -432,8 +435,8 @@ def _most_common_line_ending(s: str) -> str:
 
 def fix_file_contents(
         contents: str,
-        imports_to_add: Tuple[str, ...] = (),
-        imports_to_remove: Tuple[str, ...] = (),
+        imports_to_add: tuple[str, ...] = (),
+        imports_to_remove: tuple[str, ...] = (),
         imports_to_replace: Iterable[ImportToReplace] = (),
         **sort_kwargs: Any,
 ) -> str:
@@ -454,7 +457,7 @@ def fix_file_contents(
 
 def _fix_file(filename: str, args: argparse.Namespace) -> int:
     if filename == '-':
-        contents_bytes = getattr(sys.stdin, 'buffer', sys.stdin).read()
+        contents_bytes = sys.stdin.buffer.read()
     else:
         with open(filename, 'rb') as f:
             contents_bytes = f.read()
@@ -478,13 +481,14 @@ def _fix_file(filename: str, args: argparse.Namespace) -> int:
         unclassifiable_application_modules=args.unclassifiable,
     )
     if filename == '-':
-        print(new_contents, end='')
+        if args.diff_only:
+            _report_diff(contents, new_contents, '')
+        else:
+            print(new_contents, end='')
     elif contents != new_contents:
         if args.diff_only:
             _report_diff(contents, new_contents, filename)
         elif args.print_only:
-            print('!!! --print-only is deprecated', file=sys.stderr)
-            print('!!! maybe use `-` instead?', file=sys.stderr)
             print(f'==> {filename} <==', file=sys.stderr)
             print(new_contents, end='')
         else:
@@ -506,62 +510,150 @@ def _report_diff(contents: str, new_contents: str, filename: str) -> None:
             fromfile=filename, tofile=filename,
         ),
     )
-    if not diff.endswith('\n'):
+    if diff and not diff.endswith('\n'):
         diff += '\n\\ No newline at end of file\n'
 
     print(diff, end='')
 
 
-FUTURE_IMPORTS: Tuple[Tuple[str, Tuple[str, ...]], ...] = (
-    ('py22', ('nested_scopes',)),
-    ('py23', ('generators',)),
-    ('py26', ('with_statement',)),
-    (
-        'py3',
-        ('division', 'absolute_import', 'print_function', 'unicode_literals'),
-    ),
-    ('py37', ('generator_stop',)),
-)
-BUILTINS = (  # from python-future
-    'ascii', 'bytes', 'chr', 'dict', 'filter', 'hex', 'input', 'int', 'list',
-    'map', 'max', 'min', 'next', 'object', 'oct', 'open', 'pow', 'range',
-    'round', 'str', 'super', 'zip',
-)
+REMOVALS: dict[tuple[int, ...], set[str]] = collections.defaultdict(set)
+REPLACES: dict[tuple[int, ...], set[str]] = collections.defaultdict(set)
 
+REMOVALS[(3,)].add('from io import open')
 
-def _add_future_options(parser: argparse.ArgumentParser) -> None:
-    prev: List[str] = []
-    for py, removals in FUTURE_IMPORTS:
-        opt = f'--{py}-plus'
-        futures = ', '.join(removals)
-        implies = f'. implies: {", ".join(prev)}' if prev else ''
-        parser.add_argument(
-            opt, action='store_true',
-            help=f'Remove obsolete future imports ({futures}){implies}',
-        )
-        prev.append(opt)
+# GENERATED VIA generate-future-info
+REMOVALS[(2, 2)].add('from __future__ import nested_scopes')
+REMOVALS[(2, 3)].add('from __future__ import generators')
+REMOVALS[(2, 6)].add('from __future__ import with_statement')
+REMOVALS[(3,)].add('from __future__ import absolute_import, division, print_function, unicode_literals')  # noqa: E501
+REMOVALS[(3, 7)].add('from __future__ import generator_stop')
+# END GENERATED
 
+# GENERATED VIA generate-typing-rewrite-info
+# Using:
+#     flake8-typing-imports==1.12.0
+#     mypy_extensions==0.4.3
+#     typing_extensions==4.0.1
+REPLACES[(3, 7)].update((
+    'mypy_extensions=typing:NoReturn',
+))
+REPLACES[(3, 8)].update((
+    'mypy_extensions=typing:TypedDict',
+))
+REPLACES[(3, 6)].update((
+    'typing_extensions=typing:AsyncIterable',
+    'typing_extensions=typing:AsyncIterator',
+    'typing_extensions=typing:Awaitable',
+    'typing_extensions=typing:ClassVar',
+    'typing_extensions=typing:ContextManager',
+    'typing_extensions=typing:Coroutine',
+    'typing_extensions=typing:DefaultDict',
+    'typing_extensions=typing:NewType',
+    'typing_extensions=typing:TYPE_CHECKING',
+    'typing_extensions=typing:Text',
+    'typing_extensions=typing:Type',
+    'typing_extensions=typing:get_type_hints',
+    'typing_extensions=typing:overload',
+))
+REPLACES[(3, 7)].update((
+    'typing_extensions=typing:AsyncContextManager',
+    'typing_extensions=typing:AsyncGenerator',
+    'typing_extensions=typing:ChainMap',
+    'typing_extensions=typing:Counter',
+    'typing_extensions=typing:Deque',
+))
+REPLACES[(3, 8)].update((
+    'typing_extensions=typing:Final',
+    'typing_extensions=typing:Literal',
+    'typing_extensions=typing:OrderedDict',
+    'typing_extensions=typing:Protocol',
+    'typing_extensions=typing:SupportsIndex',
+    'typing_extensions=typing:TypedDict',
+    'typing_extensions=typing:final',
+    'typing_extensions=typing:get_args',
+    'typing_extensions=typing:get_origin',
+    'typing_extensions=typing:runtime_checkable',
+))
+REPLACES[(3, 9)].update((
+    'typing_extensions=typing:Annotated',
+))
+REPLACES[(3, 10)].update((
+    'typing_extensions=typing:Concatenate',
+    'typing_extensions=typing:ParamSpec',
+    'typing_extensions=typing:TypeAlias',
+    'typing_extensions=typing:TypeGuard',
+))
+# END GENERATED
 
-def _version_removals(args: argparse.Namespace) -> Generator[str, None, None]:
-    implied = False
-    to_remove: List[str] = []
-    for py, removals in reversed(FUTURE_IMPORTS):
-        implied |= getattr(args, f'{py}_plus')
-        if implied:
-            to_remove.extend(removals)
-    if to_remove:
-        yield f'from __future__ import {", ".join(to_remove)}'
-
-    if _is_py3(args):
-        yield from SIX_REMOVALS
-        yield 'from io import open'
-        yield 'from builtins import *'
-        yield f'from builtins import {", ".join(BUILTINS)}'
+# GENERATED VIA generate-typing-pep585-rewrites
+REPLACES[(3, 9)].update((
+    'typing=collections.abc:AsyncGenerator',
+    'typing=collections.abc:AsyncIterable',
+    'typing=collections.abc:AsyncIterator',
+    'typing=collections.abc:Awaitable',
+    'typing=collections.abc:ByteString',
+    'typing=collections.abc:Callable',
+    'typing=collections.abc:Collection',
+    'typing=collections.abc:Container',
+    'typing=collections.abc:Coroutine',
+    'typing=collections.abc:Generator',
+    'typing=collections.abc:Hashable',
+    'typing=collections.abc:ItemsView',
+    'typing=collections.abc:Iterable',
+    'typing=collections.abc:Iterator',
+    'typing=collections.abc:KeysView',
+    'typing=collections.abc:Mapping',
+    'typing=collections.abc:MappingView',
+    'typing=collections.abc:MutableMapping',
+    'typing=collections.abc:MutableSequence',
+    'typing=collections.abc:MutableSet',
+    'typing=collections.abc:Reversible',
+    'typing=collections.abc:Sequence',
+    'typing=collections.abc:Sized',
+    'typing=collections.abc:ValuesView',
+    'typing=collections:ChainMap',
+    'typing=collections:Counter',
+    'typing=collections:OrderedDict',
+    'typing=re:Match',
+    'typing=re:Pattern',
+    'typing.re=re:Match',
+    'typing.re=re:Pattern',
+))
+# END GENERATED
 
+# GENERATED VIA generate-python-future-info
+# Using future==0.18.2
+REMOVALS[(3,)].update((
+    'from builtins import *',
+    'from builtins import ascii',
+    'from builtins import bytes',
+    'from builtins import chr',
+    'from builtins import dict',
+    'from builtins import filter',
+    'from builtins import hex',
+    'from builtins import input',
+    'from builtins import int',
+    'from builtins import isinstance',
+    'from builtins import list',
+    'from builtins import map',
+    'from builtins import max',
+    'from builtins import min',
+    'from builtins import next',
+    'from builtins import object',
+    'from builtins import oct',
+    'from builtins import open',
+    'from builtins import pow',
+    'from builtins import range',
+    'from builtins import round',
+    'from builtins import str',
+    'from builtins import super',
+    'from builtins import zip',
+))
+# END GENERATED
 
 # GENERATED VIA generate-six-info
-# Using six==1.14.0
-SIX_REMOVALS = [
+# Using six==1.15.0
+REMOVALS[(3,)].update((
     'from six import callable',
     'from six import next',
     'from six.moves import filter',
@@ -569,8 +661,8 @@ SIX_REMOVALS = [
     'from six.moves import map',
     'from six.moves import range',
     'from six.moves import zip',
-]
-SIX_RENAMES = [
+))
+REPLACES[(3,)].update((
     'six.moves.BaseHTTPServer=http.server',
     'six.moves.CGIHTTPServer=http.server',
     'six.moves.SimpleHTTPServer=http.server',
@@ -635,23 +727,56 @@ SIX_RENAMES = [
     'six=functools:wraps',
     'six=io:BytesIO',
     'six=io:StringIO',
-]
+))
+# END GENERATED
+
+# GENERATED VIA generate-mock-info
+# Using mock==4.0.3
+REPLACES[(3,)].update((
+    'mock.mock=unittest.mock:ANY',
+    'mock.mock=unittest.mock:DEFAULT',
+    'mock.mock=unittest.mock:FILTER_DIR',
+    'mock.mock=unittest.mock:MagicMock',
+    'mock.mock=unittest.mock:Mock',
+    'mock.mock=unittest.mock:NonCallableMagicMock',
+    'mock.mock=unittest.mock:NonCallableMock',
+    'mock.mock=unittest.mock:PropertyMock',
+    'mock.mock=unittest.mock:call',
+    'mock.mock=unittest.mock:create_autospec',
+    'mock.mock=unittest.mock:mock_open',
+    'mock.mock=unittest.mock:patch',
+    'mock.mock=unittest.mock:sentinel',
+    'mock=unittest.mock:ANY',
+    'mock=unittest.mock:DEFAULT',
+    'mock=unittest.mock:FILTER_DIR',
+    'mock=unittest.mock:MagicMock',
+    'mock=unittest.mock:Mock',
+    'mock=unittest.mock:NonCallableMagicMock',
+    'mock=unittest.mock:NonCallableMock',
+    'mock=unittest.mock:PropertyMock',
+    'mock=unittest.mock:call',
+    'mock=unittest.mock:create_autospec',
+    'mock=unittest.mock:mock_open',
+    'mock=unittest.mock:patch',
+    'mock=unittest.mock:sentinel',
+))
 # END GENERATED
 
 
-def _is_py3(args: argparse.Namespace) -> bool:
-    for py, _ in FUTURE_IMPORTS:
-        if py.startswith('py3') and getattr(args, f'{py}_plus'):
-            return True
-    else:
-        return False
+def _add_version_options(parser: argparse.ArgumentParser) -> None:
+    versions = sorted(REMOVALS.keys() | REPLACES.keys())
 
-
-def _six_replaces(args: argparse.Namespace) -> List[ImportToReplace]:
-    if _is_py3(args):
-        return [_validate_replace_import(s) for s in SIX_RENAMES]
-    else:
-        return []
+    msg = 'Removes/updates obsolete imports; implies all older versions.'
+    parser.add_argument(
+        f'--py{"".join(str(n) for n in versions[0])}-plus', help=msg,
+        action='store_const', dest='min_version', const=versions[0],
+        default=(0,),
+    )
+    for version in versions[1:]:
+        parser.add_argument(
+            f'--py{"".join(str(n) for n in version)}-plus', help=msg,
+            action='store_const', dest='min_version', const=version,
+        )
 
 
 def _validate_import(s: str) -> str:
@@ -675,19 +800,23 @@ def _validate_replace_import(s: str) -> ImportToReplace:
         return orig_mod.split('.'), new_mod.split('.'), attr
 
 
-def main(argv: Optional[Sequence[str]] = None) -> int:
+def main(argv: Sequence[str] | None = None) -> int:
     parser = argparse.ArgumentParser()
-    parser.add_argument('filenames', nargs='*')
+    parser.add_argument(
+        'filenames', nargs='*',
+        help='If `-` is given, reads from stdin and writes to stdout.',
+    )
     group = parser.add_mutually_exclusive_group(required=False)
     group.add_argument(
         '--diff-only', action='store_true',
-        help='Show unified diff instead of applying reordering.',
+        help='(Deprecated) Show unified diff instead of applying reordering.',
     )
     group.add_argument(
         '--print-only', action='store_true',
         help=(
             '(Deprecated) '
-            'Print the output of a single file after reordering.'
+            'Print the output of a single file after reordering. '
+            'Consider using `-` for the filename instead.'
         ),
     )
     parser.add_argument('--exit-zero-even-if-changed', action='store_true')
@@ -734,24 +863,45 @@ def main(argv: Optional[Sequence[str]] = None) -> int:
     parser.add_argument(
         '--separate-relative', action='store_true',
         help=(
-            'Separate explicit relative (`from . import ...`) imports into a '
-            'separate block.'
+            '(Deprecated) Separate explicit relative (`from . import ...`) '
+            'imports into a separate block.'
         ),
     )
 
     parser.add_argument(
         '--separate-from-import', action='store_true',
         help=(
-            'Separate `from xx import xx` imports from `import xx` imports'
-            ' with a new line .'
+            '(Deprecated) Separate `from xx import xx` imports from '
+            '`import xx` imports with a new line.'
         ),
     )
 
-    _add_future_options(parser)
+    _add_version_options(parser)
 
     args = parser.parse_args(argv)
-    args.remove_import.extend(_version_removals(args))
-    args.replace_import.extend(_six_replaces(args))
+
+    for option in (
+        'diff_only',
+        'print_only',
+        'separate_relative',
+        'separate_from_import',
+    ):
+        if getattr(args, option):
+            print(
+                f'warning: --{option.replace("_", "-")} is deprecated '
+                f'and will be removed',
+                file=sys.stderr,
+            )
+
+    for k, v in REMOVALS.items():
+        if args.min_version >= k:
+            args.remove_import.extend(v)
+
+    for k, v in REPLACES.items():
+        if args.min_version >= k:
+            args.replace_import.extend(
+                _validate_replace_import(replace_s) for replace_s in v
+            )
 
     if os.environ.get('PYTHONPATH'):
         sys.stderr.write('$PYTHONPATH set, import order may be unexpected\n')
@@ -764,4 +914,4 @@ def main(argv: Optional[Sequence[str]] = None) -> int:
 
 
 if __name__ == '__main__':
-    exit(main())
+    raise SystemExit(main())
